import logging

import grpc

from ..utils import get_is_async

log = logging.getLogger(__name__)


class GRPCInterface:
    """Represents a generic gRPC interface for making requests to a gRPC server.

    This class is designed to be subclassed for specific gRPC services, providing
    a common interface for making synchronous and asynchronous requests.
    """

    stub = NotImplemented

    def __init__(
        self, app_key: str, uri: str, is_async: bool = False, timeout: int = 7
    ):
        self.app_key = app_key
        self.uri = uri
        self._is_async = get_is_async(is_async)
        self.timeout = timeout

    def make_request(self, stub_call, request, *args, **kwargs):
        try:
            with grpc.insecure_channel(self.uri) as channel:
                stub = self.stub(channel)
                response = getattr(stub, stub_call)(request, timeout=self.timeout)
                return self.process_response(stub_call, response, *args, **kwargs)
        except Exception as e:
            log.exception(f"Error making {self.__class__.__name__} request: {e}")
            return None

    async def make_request_async(self, stub_call, request, *args, **kwargs):
        try:
            async with grpc.aio.insecure_channel(self.uri) as channel:
                stub = self.stub(channel)
                response = await getattr(stub, stub_call)(request, timeout=self.timeout)
                return self.process_response(stub_call, response, *args, **kwargs)
        except Exception as e:
            log.exception(f"Error making {self.__class__.__name__} request: {e}")
            return None

    def process_response(self, stub_call: str, response, *args, **kwargs):
        if response is None:
            logging.warning(
                f"Error processing response for {self.__class__.__name__}.{stub_call}: response was None"
            )
            return None

        if response.response_header.success is False:
            # fixme: some classes (modbus, dda) have response_message instead of message (e.g camera, platform)
            message = getattr(response.response_header, "message", None) or getattr(
                response.response_header, "response_message", None
            )
            log.warning(
                f"Error processing {stub_call} response "
                f"({self.__class__.__name__}.{stub_call}): {message}"
            )
            return None

        return response



from .utils import (
    map_reading as map_reading,
    find_object_with_key as find_object_with_key,
    find_path_to_key as find_path_to_key,
    get_is_async as get_is_async,
    maybe_async as maybe_async,
    wrap_try_except as wrap_try_except,
    wrap_try_except_async as wrap_try_except_async,
    call_maybe_async as call_maybe_async,
    on_change as on_change,
    setup_logging as setup_logging,
    CaseInsensitiveDict as CaseInsensitiveDict,
    CaseInsensitiveDictEncoder as CaseInsensitiveDictEncoder,
    LogFormatter as LogFormatter,
)
from .kalman import (
    apply_kalman_filter as apply_kalman_filter,
    apply_async_kalman_filter as apply_async_kalman_filter,
)
from .pid import PID as PID
from .deprecator import deprecated as deprecated
from .diff import (
    apply_diff as apply_diff,
    generate_diff as generate_diff,
    maybe_load_json as maybe_load_json,
)

from .alarm import create_alarm as create_alarm


import json
import time
import csv
from datetime import datetime
from os import PathLike
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pydoover.cloud.api import Client


class Message:
    """Represents a message in a channel.

    Attributes
    ----------
    id: str
        The unique identifier for the message.
    channel_id: str
        The unique identifier for the channel the message belongs to.
    agent_id: str
        The unique identifier for the agent that sent the message.
    channel_name: str
        The name of the channel the message belongs to.
    """

    def __init__(
        self,
        client: "Client",
        data: dict | str,
        channel_id: str = None,
        agent_id: str = None,
        channel_name: str = None,
    ):
        self.id = None
        self._timestamp: float = None

        self.client = client
        self.channel_id = channel_id
        self.agent_id = agent_id
        self.channel_name = channel_name
        self._payload = None

        if data is not None:
            self._from_data(data)

    def __repr__(self):
        if self._payload is not None:
            return f"<Message message_id={self.id}, payload={self._payload}>"
        return f"<Message message_id={self.id}>"

    def _from_data(self, data: dict[str, Any]):
        # {'agent': '9fb5d629-ce7f-4b08-b17a-c267cbcd0427', 'message': 'a7b493dd-4577-4f81-ac3d-f1b3be680b12', 'type': 'base', 'timestamp': 1715646840.250541}
        self.id = data.get("message", None)
        self.agent_id = data.get("agent", None)
        self.channel_name = data.get("channel_name", None)
        self._timestamp = data.get("timestamp", None)

        if not self.channel_id:
            self.channel_id = data.get("channel")

        self._payload = data.get("payload")

    def to_dict(self):
        return {
            "message": self.id,
            "agent": self.agent_id,
            "timestamp": self.timestamp,
            "channel": self.channel_id,
            "channel_name": self.channel_name,
            "payload": self._payload,
        }

    def update(self) -> None:
        """Fetches the latest data for the message from the server and updates the instance attributes."""
        data = self.client._get_message_raw(self.channel_id, self.id)
        self._from_data(data)

    def delete(self) -> None:
        """Deletes the message from the channel."""
        self.client._delete_message_raw(self.channel_id, self.id)

    def fetch_payload(self) -> dict | str:
        """Fetches the payload of the message from the site."""
        if self._payload is not None:
            return self._payload

        data = self.client._get_message_raw(self.channel_id, self.id)
        self._payload = json.loads(data["payload"])
        return self._payload

    @property
    def age(self) -> float:
        """Returns the age of the message in seconds since it was created."""
        return time.time() - self._timestamp

    @property
    def timestamp(self) -> datetime:
        """Returns the timestamp of the message as a datetime object in UTC."""
        return datetime.fromtimestamp(self._timestamp)

    def get_age(self) -> float:
        return time.time() - self._timestamp

    def get_timestamp(self) -> datetime:
        return datetime.fromtimestamp(self._timestamp)

    @staticmethod
    def from_csv_export(
        client: "Client", csv_file_path: str | PathLike
    ) -> list["Message"]:
        """Create a list of Message instances from a CSV export file.

        Parameters
        ----------
        client: Client
            The client instance to use for API interactions.
        csv_file_path: str
            The path to the CSV file containing the exported messages.

        Returns
        -------
        list[Message]
            A list of Message instances created from the CSV data.
        """

        messages = []

        # Open and read the CSV file using the csv module
        with open(csv_file_path, "r", newline="") as file:
            reader = csv.DictReader(file)  # Use DictReader to handle headers

            for row in reader:
                # Extract data from the row
                key = row["Key"]
                timestamp = row["Timestamp (UTC)"]
                channel_name = row["Channel"]
                channel_id = row["Channel ID"]
                # agent_name = row["Agent"]
                agent_id = row["Agent ID"]
                payload = row["Payload"]

                # Convert timestamp to UTC epoch timestamp
                timestamp = datetime.fromisoformat(timestamp).timestamp()

                # Create a Message instance
                message = Message(
                    client,
                    data=None,
                    channel_id=channel_id,
                    agent_id=agent_id,
                    channel_name=channel_name,
                )

                message.id = key
                message._timestamp = timestamp
                message._payload = json.loads(payload)

                messages.append(message)

        # Sort the messages by timestamp
        messages.sort(key=lambda x: x.timestamp)

        return messages



import base64
import os
import shutil
import uuid
import mimetypes
import logging
import sys
import importlib
from datetime import datetime
from os import PathLike

from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from .client import Client
    from .agent import Agent
    from .message import Message


class Channel:
    """Base class for Doover channels.

    Channels are used to store state between applications, UI, tasks, processors and more in the Doover system.


    .. container:: operations

        .. describe:: x == y

            Checks if two channels are equal.



    Attributes
    ----------
    key: str
        The unique identifier for the channel.
    name: str
        The name of the channel.
    agent_id: str
        The agent that owns the channel.
    """

    def __init__(self, *, client, data):
        self.client: "Client" = client
        self._aggregate = None
        self._messages = None

        self._from_data(data)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def _from_data(self, data):
        self.id = data["channel"]
        self.key = data["channel"]
        self.name = data["name"]
        # from the get_agent endpoint this is `agent`, from the get_channel endpoint this is `owner`.
        self.agent_id = data.get("owner") or data.get("agent")
        self._agent = None

        try:
            self._aggregate = data["aggregate"]["payload"]
        except KeyError:
            self._aggregate = None

    @property
    def aggregate(self) -> Any:
        """Returns the aggregate data for the channel."""
        return self._aggregate

    def update(self):
        """Updates the channel data from the server."""
        res = self.client._get_channel_raw(self.id)
        self._from_data(res)

    def get_tunnel_url(self, address: str) -> str | None:
        """Returns the tunnel URL for a given address if it exists in the channel's aggregate.

        Parameters
        ----------
        address : str
            The address to look for in the tunnels.

        Returns
        -------
        str or None
            The tunnel URL if found, otherwise None.

        Raises
        -------
        RuntimeError
            If the channel name is not "tunnels".
        """
        if self.name != "tunnels":
            raise RuntimeError("Tunnels are only valid in the `tunnels` channel.")

        agg = self.fetch_aggregate()
        try:
            tunnels = agg["open"]
        except KeyError:
            return

        found = [t for t in tunnels if t["address"] == address]
        if found:
            return found[0]["url"]

    def fetch_agent(self) -> "Agent":
        """Fetches the agent that owns this channel.

        Returns
        -------
        :class:`pydoover.cloud.api.Agent`
            The agent object that owns this channel.

        Raises
        ------
        :class:`pydoover.cloud.api.NotFound`
            If the agent with the specified ID does not exist or you don't have permission to view it.
        """
        if self._agent is not None:
            return self._agent

        self._agent = self.client.get_agent(self.agent_id)
        return self._agent

    def fetch_aggregate(self) -> dict | str:
        """Fetches the aggregate data for the channel.

        Returns
        -------
        Any
            The aggregate data for the channel, or None if not available.
            This is generally a dictionary for JSON channels, or a string for text channels.
        """
        if self._aggregate is not None:
            return self._aggregate

        self.update()
        return self._aggregate

    def fetch_messages(self, num_messages: int = 10) -> list["Message"]:
        """Fetches the latest messages from the channel.

        Parameters
        ----------
        num_messages : int, optional
            The number of messages to fetch, by default 10.

        Returns
        -------
        list[Message]
            A list of messages from the channel

        """
        if self._messages is not None and len(self._messages) >= num_messages:
            return self._messages

        self._messages = self.client.get_channel_messages(
            self.id, num_messages=num_messages
        )
        return self._messages

    def fetch_messages_in_window(
        self, window_start: datetime, window_end: datetime
    ) -> list["Message"]:
        """Fetches messages from the channel within a specific time window.

        Parameters
        ----------
        window_start : datetime
            The start of the time window.
        window_end : datetime
            The end of the time window.

        Returns
        -------
        list[Message]
            A list of messages from the channel within the specified time window.
        """

        return self.client.get_channel_messages_in_window(
            self.id, window_start, window_end
        )

    def publish(
        self,
        data: Any,
        save_log: bool = True,
        log_aggregate: bool = False,
        override_aggregate: bool = False,
        timestamp: Optional[datetime] = None,
    ):
        """Publishes data to the channel.

        Parameters
        ----------
        data : Any
            The data to publish to the channel. This can be any JSON-serializable object.
        save_log : bool, optional
            Whether to save the published data in the channel's log, by default True.
        log_aggregate : bool, optional
            Whether to log the aggregate data after publishing, by default False.
        override_aggregate : bool, optional
            Whether to override the existing aggregate data with the published data, by default False.
        timestamp : datetime, optional
            The timestamp to use for the published data. If not provided, the current time is used.

        """
        return self.client.publish_to_channel(
            self.id, data, save_log, log_aggregate, override_aggregate, timestamp
        )

    @property
    def last_message(self) -> Optional["Message"]:
        """:class:`pydoover.cloud.api.Message` : Returns the last message published to the channel."""
        messages = self.fetch_messages(num_messages=1)
        if messages is None or len(messages) == 0:
            return None
        return messages[0]

    @property
    def last_update_age(self) -> float | None:
        """Returns the age of the last message in seconds, or None if there are no messages."""
        last_message = self.last_message
        if last_message is None:
            return None
        return last_message.get_age()

    def update_from_file(
        self, file_path: str | PathLike, mime_type: str = None
    ) -> None:
        """Updates the channel with the contents of a file.

        Parameters
        ----------
        file_path : str or PathLike
            The path to the file to read and publish.
        mime_type : str, optional
            The MIME type of the file. If not provided, it will be guessed based on the file extension.
        """

        if mime_type is None:
            mime_type, _ = mimetypes.guess_type(file_path)
            # mime_type = "application/octet-stream"

        with open(file_path, "rb") as f:
            b64_data = base64.b64encode(f.read()).decode()

        msg = {"output_type": mime_type, "output": b64_data}
        self.publish(msg)


class Processor(Channel):
    """Represents a Doover processor channel.

    Processors are lambda-like functions which run on a trigger.

    A processor channel contains a base64 encoded .zip file of the code
    required to run a processor - either a zip of the source code, or a docker-compose file.
    """

    def update_from_package(self, package_dir: PathLike) -> None:
        """Updates the processor channel with a package directory.

        This method zips the contents of the package directory and publishes it to the channel.

        Parameters
        ----------
        package_dir : str or PathLike
            The path to the directory containing the package files.
        """
        fp = f"/tmp/{uuid.uuid4()}"
        shutil.make_archive(fp, "zip", package_dir)

        with open(f"{fp}.zip", "rb") as f:
            zip_bytes = f.read()
            b64_package = base64.b64encode(zip_bytes).decode()

        self.publish(b64_package)
        os.remove(f"{fp}.zip")

    def invoke_locally(
        self,
        package_dir: str | PathLike,
        agent_id: str,
        access_token: str,
        api_endpoint: str = "https://my.doover.dev",
        package_config: dict = None,
        msg_obj: dict = None,
        task_id: str = None,
        log_channel: str = None,
        agent_settings: dict = None,
    ):
        """Invokes the processor locally with the given parameters.

        This method loads the processor code from the specified package directory and executes it.

        This is very similar to the execution process called from the `doover-task-sandbox`

        Parameters
        ----------
        package_dir : str or PathLike
            The path to the directory containing the processor code.
        agent_id : str
            The agent ID to assume to be the owner of this invocation  / channel
        access_token : str
            The access token to use for API interactions.
        api_endpoint : str, optional
            The API endpoint to use for interactions, by default "https://my.doover.dev".
        package_config : dict, optional
            The config for the trigger
        msg_obj : dict, optional
            The message object that triggered this processor, by default an empty dictionary.
        task_id : str, optional
            The ID of the task channel that invoked this processor, by default None.
        log_channel : str, optional
            The key of the channel to publish logs to, by default None.
        agent_settings : dict, optional
            Settings for the agent, including deployment configuration, by default an empty dictionary.
        """
        package_config = package_config or {}
        msg_obj = msg_obj or {}
        agent_settings = agent_settings or {}

        logging.basicConfig(level=logging.DEBUG)
        sys.path.append(package_dir)

        # Construct the full path to "target.py" within package_dir
        target_path = os.path.join(package_dir, "target.py")

        ## import the loaded generator file
        spec = importlib.util.spec_from_file_location("target", target_path)
        target_task = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(target_task)

        # from .target import generator
        target_task = getattr(target_task, "target")

        #     'agent_id' : The Doover agent id invoking the task e.g. '9843b273-6580-4520-bdb0-0afb7bfec049'
        #     'access_token' : A temporary token that can be used to interact with the Doover API .e.g 'ABCDEFGHJKLMNOPQRSTUVWXYZ123456890',
        #     'api_endpoint' : The API endpoint to interact with e.g. "https://my.doover.com",
        #     'package_config' : A dictionary object with configuration for the task - as stored in the task channel in Doover,
        #     'msg_obj' : A dictionary object of the msg that has invoked this task,
        #     'task_id' : The identifier string of the task channel used to run this processor,
        #     'log_channel' : The identifier string of the channel to publish any logs to
        #     'agent_settings' : {
        #       'deployment_config' : {} # a dictionary of the deployment config for this agent
        task_obj = target_task(
            agent_id=agent_id,
            access_token=access_token,
            api_endpoint=api_endpoint,
            package_config=package_config,
            msg_obj=msg_obj,
            task_id=task_id,
            log_channel=log_channel,
            agent_settings=agent_settings,
            # *args, **kwargs,
        )

        task_obj.execute()


class Task(Channel):
    """Represents a Doover task channel.

    Tasks are channels that trigger a processor when a message is published to them.

    Attributes
    ----------
    processor_key : str
        The key of the processor that this task triggers.
    """

    def _from_data(self, data):
        super()._from_data(data)
        self.processor_key: str = data.get("processor")
        self.processor_id = self.processor_key
        self._processor: Processor = None

    def fetch_processor(self) -> Processor | None:
        """Fetches the processor channel associated with this task.

        Returns
        -------
        Processor or None
            The processor channel that this task triggers, or None if there is no processor attached.
        """
        if self._processor is not None:
            return self._processor
        if self.processor_id is None:
            return

        self._processor = self.client.get_channel(self.processor_id)
        return self._processor

    def subscribe_to_channel(self, channel_id: str):
        """Subscribes this task to a channel.

        Parameters
        ----------
        channel_id : str
            The channel to subscribe this task to.
        """
        return self.client.subscribe_to_channel(channel_id, self.id)

    def unsubscribe_from_channel(self, channel_id: str):
        """Unsubscribes this task from a channel.

        Parameters
        ----------
        channel_id : str
            The channel to unsubscribe this task from.
        """
        return self.client.unsubscribe_from_channel(channel_id, self.id)

    def invoke_locally(
        self,
        package_dir: str | PathLike,
        msg_obj: dict | str,
        agent_settings: dict,
        agent_id: str = None,
    ):
        """Invokes the task locally with the given parameters.

        This method loads the processor code from the specified package directory and executes it.

        Parameters
        ----------
        package_dir : str or PathLike
            The path to the directory containing the processor code.
        msg_obj : dict or str
            The message to invoke the processor with
        agent_settings : dict
            Settings for the agent, including deployment configuration.
        agent_id : str, optional
            The agent ID to assume to be the owner of this invocation / channel, by default None.
        """
        processor = self.fetch_processor()
        if processor is None:
            return

        agent_id = agent_id or self.client.agent_id
        access_token = self.client.access_token.token
        api_endpoint = self.client.base_url
        package_config = self.fetch_aggregate()
        task_id = self.id

        log_channel = None

        processor.invoke_locally(
            package_dir,
            agent_id,
            access_token,
            api_endpoint,
            package_config,
            msg_obj,
            task_id,
            log_channel,
            agent_settings,
        )



import copy
import json
import logging
import pathlib
import re

from enum import EnumType, Enum as _Enum
from typing import Any

log = logging.getLogger(__name__)
KEY_VALIDATOR = re.compile(r"^[ a-zA-Z0-9_-]*$")


def transform_key(key: str):
    return key.lower().replace(" ", "_")


def check_key(key: str):
    if not KEY_VALIDATOR.match(key):
        raise ValueError(
            f"Invalid config key {key}. Keys must only contain alphanumeric characters, "
            f"hyphens (-), underscores (_) and spaces ( )."
        )


class NotSet:
    pass


class Schema:
    """Represents the configuration schema for a Doover application.

    A config schema is a definition of the config for your application.

    It is used as `.config` in your application and can generate a JSON Schema
    which will provide user validation and a "form" in the Doover UI.

    Any attributes added in the `__init__` method will be added to the schema.
    Order is preserved from the order you define them in the `__init__` method.

    The schema can be exported to a JSON file using the `export` method, although in a template application this is done for you.

    To export the schema to the `doover_config.json` file, use the Doover cli: ``doover config-schema export``.
    This will validate and export the schema to the `doover_config.json` file in the root of your Doover project.

    Examples
    --------

    >>> from pydoover import config
    >>> class MyAppConfig(config.Schema):
    ...     def __init__(self):
    ...         self.pump_pin = config.Integer("Digital Output Number", description="The digital output pin to drive the pump.")
    ...         self.pump_on_time = config.Number("Pump On Time", default=5.2, description="The time in seconds to run the pump.")
    ...         self.engine_type = config.Enum(
    ...             "Engine Type",
    ...             choices=["Honda", "John Deere", "Cat"],
    ...             description="The type of diesel engine attached to the pump.",
    ...         )

    """

    __element_map: "dict[str, ConfigElement]" = {}

    def add_element(self, element):
        try:
            # do this here so we don't have to override __init__
            elem_map = self.__element_map
        except AttributeError:
            # this is the first element, so create the map
            elem_map = self.__element_map = dict()

        element._name = transform_key(element.display_name)
        if element._name in elem_map:
            raise ValueError(f"Duplicate element name {element._name} not allowed.")

        elem_map[element._name] = element

    def __setattr__(self, key, value):
        if isinstance(value, ConfigElement):
            # value._name = key
            self.add_element(value)
        super().__setattr__(key, value)

    def to_dict(self):
        return {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "$id": "",
            "title": "Application Config",
            "type": "object",
            "properties": {
                name: element.to_dict()
                for name, element in self.__element_map.items()
                if isinstance(element, ConfigElement)
            },
            "additionalElements": True,
            "required": [
                name
                for name, element in self.__element_map.items()
                if isinstance(element, ConfigElement) and element.required
            ],
        }

    def _inject_deployment_config(self, config: dict[str, Any]):
        for name, value in config.items():
            try:
                elem = self.__element_map[name]
            except KeyError:
                log.info(f"Skipping unknown config key {name} ({value})")
            else:
                elem.load_data(value)

        for elem_name in set(self.__element_map.keys()) - set(config.keys()):
            # catch missing required elements, and set any other elements to their default value
            elem = self.__element_map[elem_name]
            if elem.required:
                raise ValueError(
                    f"Required config element {elem_name} not found in deployment config."
                )
            elem.load_data(elem.default)

    def export(self, fp: pathlib.Path, app_name: str):
        """Export the config schema to a JSON file.

        This will export the config schema to the ``config_schema`` field in the ``doover_config.json`` file in the root of your project .

        Examples
        --------

        Generally, this will be done in the application template.

        >>> from pydoover import config
        >>> class MyAppConfig(config.Schema):
        ...     def __init__(self):
        ...         self.pump_pin = config.Integer("Digital Output Number", description="The digital output pin to drive the pump.")
        ...
        ... if __name__ == "__main__":
        ...     from pathlib import Path
        ...     MyAppConfig().export(pathlib.Path("/path/to/my/app/doover_config.json"), "my_app_name")


        Parameters
        ----------
        fp: pathlib.Path
            The path to the JSON file to export the config schema to.
        app_name: str
            The name of the application to export the config schema for.
            This will be used as the key in the `doover_config.json` file.
        """
        if fp.exists():
            data = json.loads(fp.read_text())
        else:
            data = {}

        try:
            data[app_name]["config_schema"] = self.to_dict()
        except KeyError:
            data[app_name] = {"config_schema": self.to_dict()}

        fp.write_text(json.dumps(data, indent=4))


class ConfigElement:
    """Represents a config element in the Doover configuration schema.

    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    default: int
        The default value for the integer. If NotSet, the value is required.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    """

    _type = "unknown"

    def __init__(
        self,
        display_name,
        *,
        default: Any = NotSet,
        description: str = None,
        deprecated: bool = None,
        hidden: bool = False,
    ):
        self._name = transform_key(display_name)
        self.display_name = display_name
        self.default = default
        self.description = description
        self.hidden = hidden
        self.deprecated = deprecated

        self._value = NotSet

        if (
            default is not NotSet
            and not isinstance(default, Variable)
            and default is not None
        ):
            match self._type:
                case "integer":
                    assert isinstance(default, int)
                case "number":
                    assert isinstance(default, float)
                case "string":
                    assert isinstance(default, str)
                case "boolean":
                    assert isinstance(default, bool)
                case ("array", "object"):
                    if default is not None:
                        raise ValueError(
                            "You cannot set default values for arrays and objects. It's confusing."
                        )

    @property
    def required(self):
        """Whether the config element is required."""
        return self.default is NotSet

    @property
    def value(self):
        """The value of the config element."""
        if self._value is NotSet:
            raise ValueError(f"Value for {self._name} not set. Check your config file?")
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def to_dict(self):
        payload = {
            "title": self.display_name,
            "x-name": self._name,
            "x-hidden": self.hidden,
        }

        if self._type is not None:
            payload["type"] = self._type

        if self.description is not None:
            payload["description"] = self.description

        if isinstance(self.default, Variable):
            payload["default"] = str(self.default)
        elif self.default is not NotSet:
            payload["default"] = self.default

        if self.deprecated is not None:
            payload["deprecated"] = self.deprecated

        return payload

    def load_data(self, data):
        self.value = data


class Integer(ConfigElement):
    """Represents a JSON Integer type. Internally represented as an int.

    Attributes
    -----------
    display_name: str
        The display name of the config element. This is used in the UI.
    default: int
        The default value for the integer. If NotSet, the value is required.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    minimum: int | None
        The minimum value for the integer. If None, no minimum is enforced.
    exclusive_minimum: int | None
        The exclusive minimum value for the integer. If None, no exclusive minimum is enforced.
    maximum: int | None
        The maximum value for the integer. If None, no maximum is enforced.
    exclusive_maximum: int | None
        The exclusive maximum value for the integer. If None, no exclusive maximum is enforced.
    multiple_of: int | None
        The value that the integer must be a multiple of. If None, no multiple is enforced.
    """

    _type = "integer"
    value: int

    def __init__(
        self,
        display_name,
        *,
        minimum: int = None,
        exclusive_minimum: int = None,
        maximum: int = None,
        exclusive_maximum: int = None,
        multiple_of: int = None,
        **kwargs,
    ):
        super().__init__(display_name, **kwargs)
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.multiple_of = multiple_of

    def to_dict(self):
        res = super().to_dict()
        if self.minimum is not None:
            res["minimum"] = self.minimum
        if self.exclusive_minimum is not None:
            res["exclusiveMinimum"] = self.exclusive_minimum
        if self.maximum is not None:
            res["maximum"] = self.maximum
        if self.exclusive_maximum is not None:
            res["exclusiveMaximum"] = self.exclusive_maximum
        if self.multiple_of is not None:
            res["multipleOf"] = self.multiple_of

        return res


class Number(Integer):
    """Represents a JSON Number type, for any numeric type. Internally represented as a float.

    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    default: float
        The default value for the integer. If NotSet, the value is required.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    """

    _type = "number"
    value: float


class Boolean(ConfigElement):
    """Represents a JSON Boolean type. Internally represented as a bool.

    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    default: bool
        The default value for the integer. If NotSet, the value is required.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    """

    _type = "boolean"
    value: bool


class String(ConfigElement):
    """Represents a JSON String type. Internally represented as a str.

    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    default: str
        The default value for the integer. If NotSet, the value is required.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    length: int | None
        The length of the string. If None, no length is enforced.
    pattern: str | None
        A regex pattern that the string must match. If None, no pattern is enforced.
    """

    _type = "string"
    value: str

    def __init__(
        self, display_name, *, length: int = None, pattern: str = None, **kwargs
    ):
        super().__init__(display_name, **kwargs)
        self.length = length
        self.pattern = pattern

    def to_dict(self):
        res = super().to_dict()
        if self.length is not None:
            res["length"] = self.length
        if self.pattern is not None:
            res["pattern"] = self.pattern

        return res


class Enum(ConfigElement):
    """Represents a JSON Enum type. Internally represented as a list of choices.

    The UI renders this as a drop-down.

    Examples
    --------

    You can specify a list of choices as strings or floats, or use an EnumType::

        from pydoover import config

        class MyChoice(enum.Enum):
            A = "Choice 1"
            B = "Choice 2"
            C = "Choice 3"

        class AppConfig(config.Schema):
            def __init__(self):
                self.choice = config.Enum(
                    "Choose Something",
                    choices=MyChoice,
                    default=MyChoice.A,
                )

                self.other_choice = config.Enum(
                    "Other Choice",
                    choices=["a", "b", "c"],
                    default="a"
                )


    You can also set enum values to be objects to allow for custom attributes, provided your object implements ``__str__``::

        from pydoover import config

        class Choice:
            def __init__(self, name, level):
                self.name = name
                self.level = level

            def __str__(self):
                return self.name

        class ChoiceType(enum.Enum):
            A = Choice("A", 1)
            B = Choice("B", 2)
            C = Choice("C", 3)

        class AppConfig(config.Schema):
            def __init__(self):
                self.choice = config.Enum(
                    "Choose Something",
                    choices=ChoiceType,
                    default=ChoiceType.A,
                )

            @property
            def choice_value(self):
                return self.choice.value.level



    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    default: same type as choices.
        The default value for the integer. If NotSet, the value is required.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    choices: EnumType or list of str | float
        A list of choices for the enum. All choices must be of the same type (str or float).
        This optionally accepts an EnumType, with the value of the enum denoting the choice.
        The value can be an object which implements the ``__str__`` method. Each ``__str__`` value must be unique.
    """

    _type = None

    def __init__(
        self, display_name, *, choices: list | EnumType = None, default: Any, **kwargs
    ):
        if isinstance(default, _Enum):
            default = str(default.value)

        super().__init__(display_name, default=default, **kwargs)

        if isinstance(choices, EnumType):
            choices = [choice.value for choice in choices]
            self._enum_lookup = {str(choice): choice for choice in choices}
            choices = list(self._enum_lookup.keys())
        else:
            self._enum_lookup = None

        if all(isinstance(choice, str) for choice in choices):
            self._type = "string"
        elif all(isinstance(choice, float) for choice in choices):
            self._type = "number"

        self.choices = choices

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, value):
        if self._enum_lookup is None:
            self._value = value
        else:
            self._value = self._enum_lookup[value]

    def to_dict(self):
        return {
            "enum": self.choices,
            **super().to_dict(),
        }


class Array(ConfigElement):
    """Represents a JSON Array type. Internally represented as a list.

    Only a subset of JSON Schema is supported:
    - Item type
    - Minimum and maximum number of items
    - Unique items

    .. note::

        The ``default`` value is not allowed for Array elements, as it is confusing.

    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    element: ConfigElement
        The type of elements in the array. This can be any ConfigElement, such as String, Integer, etc.
    min_items: int | None
        The minimum number of items in the array. If None, no minimum is enforced.
    max_items: int | None
        The maximum number of items in the array. If None, no maximum is enforced.
    unique_items: bool | None
        Whether the items in the array must be unique. If None, no uniqueness is enforced.

    """

    _type = "array"

    def __init__(
        self,
        display_name,
        *,
        element: ConfigElement = None,
        min_items: int = None,
        max_items: int = None,
        unique_items: bool = None,
        **kwargs,
    ):
        if element and not isinstance(element, ConfigElement):
            raise ValueError("Many element must be a ConfigElement instance")
        if "default" in kwargs:
            raise ValueError(
                "Default value not allowed for Many elements. It's confusing."
            )

        super().__init__(display_name, **kwargs)

        self.element = element or ConfigElement("unknown")
        self.min_items = min_items
        self.max_items = max_items
        self.unique_items = unique_items

        self._elements = []

    def to_dict(self):
        res = super().to_dict()
        if self.element is not None:
            res["items"] = self.element.to_dict()
        if self.min_items is not None:
            res["minItems"] = self.min_items
        if self.max_items is not None:
            res["maxItems"] = self.max_items
        if self.unique_items is not None:
            res["uniqueItems"] = self.unique_items
        return res

    @property
    def elements(self) -> list[ConfigElement]:
        return self._elements

    def load_data(self, data):
        self._elements.clear()
        for row in data:
            elem = copy.deepcopy(self.element)
            elem.load_data(row)
            self._elements.append(elem)


class Object(ConfigElement):
    """Represents a JSON Object type.

    This is a complex type that can contain multiple elements, each with its own type.
    It can also have additional elements that are not defined in the schema.

    The UI renders this as a form with fields for each element.

    Examples
    --------

    >>> from pydoover import config
    >>> class MyAppConfig(config.Schema):
    ...     def __init__(self):
    ...         self.pump = config.Object(
    ...             "Pump Settings",
    ...          )
    ...         self.pump.add_elements(
    ...             config.Integer("Digital Output Number", description="The digital output pin to drive the pump."),
    ...             config.Number("On Time", default=5.2, description="The time in seconds to run the pump."),
    ...         )


    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    additional_elements: bool | dict[str, Any]
        If True, allows additional elements that are not defined in the schema.
        If a dict, defines the schema for additional elements.
        If False, no additional elements are allowed.
    """

    _type = "object"

    def __init__(
        self,
        display_name,
        *,
        additional_elements: bool | dict[str, Any] = True,
        **kwargs,
    ):
        if "default" in kwargs:
            raise ValueError(
                "Default value not allowed for Object elements. It's confusing."
            )

        super().__init__(display_name, **kwargs)
        self._elements = {}
        self.additional_elements = additional_elements

    def __setattr__(self, key, value):
        if isinstance(value, ConfigElement):
            self.add_elements(value)
        super().__setattr__(key, value)

    def __getattr__(self, key):
        # Safety check: if _elements doesn't exist yet, it means we're still initializing
        if "_elements" in self.__dict__ and key in self._elements:
            return self._elements[key]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{key}'"
        )

    def add_elements(self, *element):
        for element in element:
            if element._name in self._elements:
                raise ValueError(f"Duplicate element name {element._name} not allowed.")
            self._elements[element._name] = element

    def to_dict(self):
        res = super().to_dict()
        res["properties"] = {
            element._name: element.to_dict() for element in self._elements.values()
        }
        res["additionalElements"] = self.additional_elements
        res["required"] = [
            elem._name for elem in self._elements.values() if elem.required is True
        ]
        return res

    def load_data(self, data):
        for name, value in data.items():
            try:
                self._elements[name].load_data(value)
            except KeyError:
                if self.additional_elements is True:
                    self._elements[name] = ConfigElement(name, default=value)
                else:
                    raise ValueError(f"Unknown element {name} in config.")


class Variable:
    """Represents a variable in the config schema.

    This is a special type of config element that is used to reference other config elements.
    It is used to create dynamic references to other config elements, such as device-specific settings.

    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    scope: str
        The scope of the variable, which is usually the application name.
    name: str
        The name of the variable, which is usually the key of the config element.

    """

    def __init__(self, scope: str, name: str):
        self._scope = transform_key(scope)
        self._name = transform_key(name)

    def __str__(self):
        return f"${self._scope}.{self._name}"


class Application(ConfigElement):
    """Represents a Doover application configuration element.

    This is used to reference other Doover applications in the configuration schema.

    This is rendered as a dropdown in the UI, allowing the user to select an installed application.

    Attributes
    ----------
    display_name: str
        The display name of the config element. This is used in the UI.
    description: str | None
        A help text for the config element.
    hidden: bool
        Whether the config element should be hidden in the UI.
    """

    _type = "string"
    value: str

    def to_dict(self):
        return {"format": "doover-application", **super().to_dict()}



from copy import copy

from pydoover import config

import pytest
from jsonschema import ValidationError, validate


class ConfigSchemaA(config.Schema):
    def __init__(self):
        self.a = config.Integer("A", default=1, minimum=0)
        self.b = config.Number("B", exclusive_minimum=0)
        self.c = config.String("C", pattern=r"^[a-zA-Z0-9_]+$")
        self.d = config.Enum("D", choices=["a", "b", "c"], default="a")
        self.e = config.Boolean("E", default=False)
        self.f = config.Array("F", element=config.Integer("F Element"))

        self.g = config.Object("G")
        self.g.a = config.Integer("G A", default=1, minimum=0)
        self.g.b = config.Number("G B", exclusive_minimum=0)


SAMPLE_CONFIG_A = {
    "a": 1,
    "b": 0.5,
    "c": "valid_string",
    "d": "a",
    "e": False,
    "f": [1, 2, 3],
    "g": {"g_a": 1, "g_b": 0.5},
}
schema = ConfigSchemaA()


class TestConfigSchemaA:
    def test_ok_schema(self):
        validate(copy(SAMPLE_CONFIG_A), schema.to_dict())

    @pytest.mark.parametrize(
        "key, value",
        [
            ("a", 1.5),
            ("b", "a string"),
            ("c", True),
            ("d", 0),
            ("e", "not_a_boolean"),
            ("f", ["a", "b", "c"]),
            ("g", {"a": 1, "b": 0.5}),
            ("g", "not-an-object"),
            ("g", ["an-array-of-string"]),
        ],
    )
    def test_invalid_type(self, key, value):
        sample_config = copy(SAMPLE_CONFIG_A)
        sample_config[key] = value
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())

    def test_enum(self):
        sample_config = copy(SAMPLE_CONFIG_A)
        sample_config["d"] = "not-a-choice"
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())

        sample_config["d"] = "c"
        validate(sample_config, schema.to_dict())

    def test_minimum(self):
        sample_config = copy(SAMPLE_CONFIG_A)
        sample_config["a"] = -1
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())

        sample_config["a"] = 0
        validate(sample_config, schema.to_dict())

    def test_exclusive_min(self):
        sample_config = copy(SAMPLE_CONFIG_A)
        sample_config["b"] = 0
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())

    def test_array(self):
        sample_config = copy(SAMPLE_CONFIG_A)
        sample_config["f"] = [1, 2, "not an int"]
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())

        sample_config["f"] = []
        validate(sample_config, schema.to_dict())

        sample_config["f"] = ["not an int"]
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())

    def test_object(self):
        sample_config = copy(SAMPLE_CONFIG_A)
        sample_config["g"]["g_a"] = "not an int"
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())

        sample_config["g"]["g_b"] = 0
        with pytest.raises(ValidationError):
            validate(sample_config, schema.to_dict())



import argparse
import inspect
import typing

from .parsers import (
    extract_description,
    extract_parameters,
    json_or_str,
    int_or_list,
    bool_or_list,
    float_or_list,
    BoolFlag,
)


class SubSection:
    def __init__(
        self, section_class, name: str = "subcommands", description: str = None
    ):
        self.name = name
        self.description = description
        self.section_class = section_class

        self.uri_remapping = None
        self.section_params = []
        self.init_signature = None

        self.setup_base_arguments()

        self.subparser = None

    def mount_sub_section(self, arg_parser):
        sub_section_parser = arg_parser.add_parser(
            name=self.name, help=self.description
        )
        sub_section_parser.set_defaults(callback=sub_section_parser.print_help)

        for param_name, kwargs in self.section_params:
            sub_section_parser.add_argument(param_name, **kwargs)

        self.subparser = sub_section_parser.add_subparsers(title="Subcommands")

        for name, func in inspect.getmembers(
            self.section_class, predicate=inspect.isfunction
        ):
            if not getattr(func, "_is_command", False):
                continue

            parser = self.subparser.add_parser(
                func._command_name,
                description=extract_description(func._command_help),
                formatter_class=argparse.RawDescriptionHelpFormatter,
            )

            # Callback to call the function on the correct class passing the correct aguments to both the interface class and the method
            def func_caller(func_to_run=func, **kwargs):
                if "uri" in kwargs:
                    if self.uri_remapping is None:
                        raise Exception("uri not remapped but has been supplied")
                    kwargs[self.uri_remapping] = kwargs["uri"]

                kwargs["app_key"] = "pydoover-cli"
                # Create the interface class
                object_instance = self.section_class(
                    **{
                        k: v
                        for k, v in kwargs.items()
                        if k in self.init_signature.parameters.keys()
                    }
                )

                # Run the method on the interface class
                result = func_to_run(
                    object_instance,
                    **{
                        k: v
                        for k, v in kwargs.items()
                        if k in inspect.signature(func_to_run).parameters.keys()
                    },
                )

                if result is not None:
                    print(result)

            parser.set_defaults(callback=func_caller)
            argspec = inspect.signature(func)
            doc_string_paramaters = extract_parameters(func._command_help)
            arg_docs = func._command_arg_docs

            for param in argspec.parameters.values():
                kwargs = {"help": arg_docs.get(param.name)}
                if param.name == "self":
                    continue
                if param.name in doc_string_paramaters:
                    kwargs["help"] = (
                        doc_string_paramaters[param.name][0]
                        + " - "
                        + doc_string_paramaters[param.name][1]
                    )
                param_name = param.name

                if param.default is not inspect.Parameter.empty:
                    kwargs["default"] = param.default
                    kwargs["required"] = False
                    param_name = "--" + param_name

                self.parse_arg_type(param, kwargs)

                try:
                    parser.add_argument(param_name, **kwargs)
                except ValueError:
                    print(
                        f"Failed to setup command {self.name}.{name} parameter type: {str(param.annotation)} is unknown, please add a handler"
                    )

            if func._command_setup_api:
                parser.add_argument(
                    "--profile", help="Config profile to use.", default="default"
                )
                parser.add_argument(
                    "--agent",
                    help="Agent query string (name or ID) to use for this request.",
                    type=str,
                    default="default",
                )

            parser.add_argument(
                "--enable-traceback",
                help=argparse.SUPPRESS,
                default=False,
                action="store_true",
            )

    def setup_base_arguments(self):
        # find all the parameters needed for the base interface
        for name, func in inspect.getmembers(
            self.section_class, predicate=inspect.isfunction
        ):
            if name != "__init__":
                continue

            self.init_signature = inspect.signature(func)
            # doc_string_paramaters = parsers.extract_parameters(func._command_help)
            arg_docs = {}
            if hasattr(func, "_command_arg_docs"):
                arg_docs = func._command_arg_docs

            for param in self.init_signature.parameters.values():
                kwargs = {"help": arg_docs.get(param.name)}
                if param.name == "self":
                    continue
                # if param.name in doc_string_paramaters:
                #     kwargs["help"] = doc_string_paramaters[param.name][0] + " - " + doc_string_paramaters[param.name][1]
                param_name = param.name

                if "uri" in param_name:
                    self.uri_remapping = param_name
                    param_name = "uri"
                elif "is_async" in param_name:
                    continue
                elif "app_key" in param_name:
                    continue

                if param.default is not inspect.Parameter.empty:
                    kwargs["default"] = param.default
                    kwargs["required"] = False
                    param_name = "--" + param_name
                else:
                    print(
                        f"Error, grcp class ({self.section_class.__name__}) has a non default __init__ paramater: {param_name}"
                    )
                    continue

                self.parse_arg_type(param, kwargs)

                self.section_params.append((param_name, kwargs))

    def parse_arg_type(self, param, kwargs):
        if param.annotation is BoolFlag:
            kwargs["action"] = (
                "store_false" if kwargs["default"] is True else "store_true"
            )
        elif param.annotation == typing.Union[int, list[int]]:
            kwargs["type"] = int_or_list
        elif param.annotation == typing.Union[bool, list[bool]]:
            kwargs["type"] = bool_or_list
        elif param.annotation == typing.Union[float, list[float]]:
            kwargs["type"] = float_or_list
        elif param.annotation == typing.Union[dict, str]:
            kwargs["type"] = json_or_str
        elif param.annotation is not inspect.Parameter.empty:
            kwargs["type"] = param.annotation
        return kwargs



from .element import (
    Element as Element,
    ConnectionType as ConnectionType,
    ConnectionInfo as ConnectionInfo,
    AlertStream as AlertStream,
    Multiplot as Multiplot,
    RemoteComponent as RemoteComponent,
)
from .interaction import (
    Interaction as Interaction,
    Action as Action,
    WarningIndicator as WarningIndicator,
    HiddenValue as HiddenValue,
    SlimCommand as SlimCommand,
    StateCommand as StateCommand,
    Slider as Slider,
    action as action,
    warning_indicator as warning_indicator,
    state_command as state_command,
    hidden_value as hidden_value,
    slider as slider,
    callback as callback,
)
from .manager import UIManager as UIManager
from .misc import (
    NotSet as NotSet,
    Colour as Colour,
    Range as Range,
    Option as Option,
    Widget as Widget,
    ApplicationVariant as ApplicationVariant,
)
from .parameter import (
    Parameter as Parameter,
    TextParameter as TextParameter,
    NumericParameter as NumericParameter,
    BooleanParameter as BooleanParameter,
    DateTimeParameter as DateTimeParameter,
    numeric_parameter as numeric_parameter,
    text_parameter as text_parameter,
    boolean_parameter as boolean_parameter,
    datetime_parameter as datetime_parameter,
)
from .submodule import (
    Container as Container,
    Submodule as Submodule,
    Application as Application,
)
from .variable import (
    Variable as Variable,
    NumericVariable as NumericVariable,
    TextVariable as TextVariable,
    BooleanVariable as BooleanVariable,
    DateTimeVariable as DateTimeVariable,
)
from .camera import Camera as Camera, CameraHistory as CameraHistory



import asyncio
import copy
import datetime
import json
import logging
import time
import sys

from collections.abc import Coroutine, Callable
from typing import Any

import grpc

from .grpc_stubs import device_agent_pb2, device_agent_pb2_grpc
from ..grpc_interface import GRPCInterface
from ...utils import apply_diff, call_maybe_async, maybe_async, maybe_load_json
from ...cli.decorators import command as cli_command

log = logging.getLogger(__name__)

MaybeAsyncCallback = (
    Callable[[str, dict[str, Any] | str], None]
    | Coroutine[[str, dict[str, Any] | str], None]
)


class DeviceAgentInterface(GRPCInterface):
    """Interface for interacting with the Device Agent gRPC service.

    Attributes
    ----------
    dda_timeout : int
        Timeout for requests to the Device Agent service.
    max_connection_attempts : int
        Maximum number of attempts to connect to the Device Agent service.
    time_between_connection_attempts : int
        Time to wait between connection attempts to the Device Agent service.

    is_dda_available : bool
        Whether the Device Agent service is available. This is set to True once a successful request has been made to the service.
    is_dda_online: bool
        Whether the Device Agent service is currently online.
    has_dda_been_online: bool
        Whether the Device Agent service has been online at least once since the interface was created.

    last_channel_message_ts : dict
        A dictionary that stores the last time a message was received from each channel.
    """

    stub = device_agent_pb2_grpc.deviceAgentStub

    def __init__(
        self,
        app_key: str,
        dda_uri: str = "127.0.0.1:50051",
        is_async: bool = None,
        dda_timeout: int = 7,
        max_conn_attempts: int = 5,
        time_between_connection_attempts: int = 10,
    ):
        super().__init__(app_key, dda_uri, is_async, dda_timeout)

        self.dda_timeout = dda_timeout
        self.max_connection_attempts = max_conn_attempts
        self.time_between_connection_attempts = time_between_connection_attempts

        self.is_dda_available = False
        self.is_dda_online = False
        self.has_dda_been_online = False
        self.agent_id = None

        # this is a list of channels that the agent interface will subscribe to,
        # and a list of callbacks that will be called when a message is received,
        # as well as the aggregate data that is received from the channel
        ## for channel in default_subscriptions:
        self._subscriptions: dict[str, list[Callable]] = {}
        self._synced_channels: dict[str, bool] = {}
        self._listeners: dict[str, asyncio.Task] = {}
        self._aggregates: dict[str, dict[str, Any] | str] = {}

        self.last_channel_message_ts = {}  # this is a dictionary of the last time a message was received from a channel

    @staticmethod
    def has_persistent_connection():
        """For the Device Agent, this always returns `True`. This method exists to provide interoperability with the API client."""
        return True

    @cli_command()
    def get_is_dda_available(self):
        return self.is_dda_available

    @cli_command()
    def get_is_dda_online(self):
        return self.is_dda_online

    @cli_command()
    def get_has_dda_been_online(self):
        return self.has_dda_been_online

    @cli_command()
    @maybe_async()
    def await_dda_available(self, timeout: int = 10):
        start_time = datetime.datetime.now()
        while (
            not self.test_dda_available()
            or (datetime.datetime.now() - start_time).seconds > timeout
        ):
            time.sleep(0.1)
        return True

    async def await_dda_available_async(self, timeout: int = 10):
        start_time = datetime.datetime.now()
        while (
            not self.test_dda_available()
            or (datetime.datetime.now() - start_time).seconds > timeout
        ):
            await asyncio.sleep(0.1)
        return True

    def add_subscription(self, channel_name: str, callback: MaybeAsyncCallback) -> None:
        """Add a subscription to a channel.

        This does not block, but will start a background task to asynchronously
        invoke the callback when a new message is received on the given channel.

        The callback may be a regular function or an async function.

        Examples
        --------

        >>> async def my_callback(name, data):
        ...     print(f"Received data on {name}: {data}")
        >>> self.device_agent.add_subscription("my_channel", my_callback)

        Parameters
        ----------
        channel_name : str
            Name of channel to subscribe to.
        callback : Callable
            A callback function that will be called when a message is received on the channel.
            This callback should accept two parameters: channel_name, aggregate_data.
            It may be a regular function or an async function.
        """
        try:
            self._subscriptions[channel_name].append(callback)
        except KeyError:
            self._subscriptions[channel_name] = [callback]
            self._listeners[channel_name] = asyncio.create_task(
                self.start_subscription_listener(channel_name)
            )

    async def start_subscription_listener(self, channel_name):
        # """Start a subscription listener for the given channel.
        #
        # You should instead use the `add_subscription` method to add a subscription with a passed callback.
        #
        # This method will continually wait for updates from a channel,
        # and re-try connections that fail after `self.time_between_connection_attempts` seconds.
        # """
        while True:
            try:
                log.debug(f"Starting subscription to channel: {channel_name}")
                await self._subscribe_receive_channel_updates(channel_name)
            except Exception as e:
                log.error(
                    f"Error starting subscription listener for {channel_name}: {e}",
                    exc_info=e,
                )
                await asyncio.sleep(self.time_between_connection_attempts)

    async def _subscribe_receive_channel_updates(self, channel_name):
        async with grpc.aio.insecure_channel(self.uri) as channel:
            channel_stream = device_agent_pb2_grpc.deviceAgentStub(
                channel
            ).GetChannelSubscription(
                device_agent_pb2.ChannelSubscriptionRequest(channel_name=channel_name)
            )
            while True:
                try:
                    response: device_agent_pb2.ChannelSubscriptionResponse = (
                        await channel_stream.read()
                    )
                    log.debug(
                        f"Received response from subscription request on {channel_name}: {str(response)[:120]}"
                    )
                    self.update_dda_status(response.response_header)
                    if not response.response_header.success:
                        raise RuntimeError(
                            f"Failed to subscribe to channel {channel_name}: {response.response_header.response_message}"
                        )

                    log.debug(
                        f"Calling callback with subscription response for {channel_name}..."
                    )
                    await self.recv_update_callback(channel_name, response)
                except StopAsyncIteration:
                    log.debug("Channel stream ended.")
                    break

    async def recv_update_callback(self, channel_name, response):
        log.debug(f"Received response from subscription request: {str(response)[:100]}")
        try:
            existing = self._aggregates[channel_name]
        except KeyError:
            data = await self.get_channel_aggregate_async(channel_name)
        else:
            diff = maybe_load_json(response.message.payload)
            log.debug(
                f"Applying diff to existing data for channel {channel_name}: {diff}"
            )
            data = copy.deepcopy(existing)
            if diff:
                data = apply_diff(data, diff)
                log.debug(f"Applied diff result ({channel_name}): {data}")
            # else:
            #     log.debug(f"No diff found for channel {channel_name}. Skipping...")
            #     return

        if data in (None, "None"):
            log.warning(f"Received empty data from channel {channel_name}")
            data = {}
            # return

        self._aggregates[channel_name] = data
        self._synced_channels[channel_name] = True
        self.last_channel_message_ts[channel_name] = datetime.datetime.now()

        log.debug(
            f"Calling {len(self._subscriptions.get(channel_name, []))} "
            f"callbacks for channel {channel_name} with data: {str(data)[:100]}"
        )
        _tasks = [
            await call_maybe_async(
                callback, channel_name, copy.deepcopy(data), as_task=True
            )
            for callback in self._subscriptions.get(channel_name, [])
        ]
        # await asyncio.gather(*tasks)

    def process_response(self, stub_call: str, response, *args, **kwargs):
        self.update_dda_status(response.response_header)
        return super().process_response(stub_call, response, *args, **kwargs)

    def update_dda_status(self, header):
        if header.success:
            self.is_dda_available = True
        else:
            self.is_dda_available = False

        if header.cloud_synced:
            self.is_dda_online = True
            if not self.has_dda_been_online:
                log.info("Device Agent is online")
            self.has_dda_been_online = True
        else:
            self.is_dda_online = False

    def is_channel_synced(self, channel_name):
        """Check if a channel is synced with DDA.

        During normal operation, this should always return `True` while DDA is active.

        It is only really useful for timing during the startup process.

        Parameters
        ----------
        channel_name : str
            Name of the channel to check.

        Returns
        -------
        bool
            True if the channel is synced, False otherwise.
        """
        if channel_name not in self._subscriptions:
            return False
        if channel_name not in self._synced_channels:
            return False
        return self._synced_channels[channel_name]

    async def wait_for_channels_sync_async(
        self, channel_names: list[str], timeout: int = 5, inter_wait: float = 0.2
    ) -> bool:
        """Wait for all specified channels to be synced with DDA.

        This is invoked internally at startup to ensure that all channels are ready before proceeding with operations that depend on them.

        You shouldn't need to use this during normal operation.

        Parameters
        ----------
        channel_names : list[str]
            List of channel names to check for sync status.
        timeout : int
            Maximum time to wait for all channels to sync, in seconds.
        inter_wait : float
            Time to wait between checks, in seconds.

        Returns
        -------
        bool
            True if all channels are synced within the timeout, False otherwise.
        """
        start_time = datetime.datetime.now()
        while not all(
            [self.is_channel_synced(channel_name) for channel_name in channel_names]
        ):
            if (datetime.datetime.now() - start_time).seconds > timeout:
                return False
            await asyncio.sleep(inter_wait)
        return True

    @cli_command()
    @maybe_async()
    def get_channel_aggregate(self, channel_name: str) -> str | dict[str, Any] | None:
        """Fetch a channel's current aggregate payload.

        .. note:: This method can be called in both sync and asynchronous contexts.

        Examples
        --------
        >>> aggregate = await self.device_agent.get_channel_aggregate("my_channel")
        >>> print(aggregate)  # This will print the aggregate data as a dictionary or string.

        Parameters
        ----------
        channel_name : str
            Name of channel to get aggregate from.

        Returns
        -------
        dict | str, optional
            Aggregate data from channel. If the request fails for any reason, this will return `None`.
        """
        log.debug(f"Getting channel aggregate for {channel_name}")
        resp = self.make_request(
            "GetChannelDetails",
            device_agent_pb2.ChannelDetailsRequest(channel_name=channel_name),
        )
        if not resp:
            return None

        return maybe_load_json(resp.channel.aggregate)

    async def get_channel_aggregate_async(self, channel_name):
        log.debug(f"Getting channel aggregate (async) for {channel_name}")
        resp = await self.make_request_async(
            "GetChannelDetails",
            device_agent_pb2.ChannelDetailsRequest(channel_name=channel_name),
        )
        if not resp:
            return

        return maybe_load_json(resp.channel.aggregate)

    @cli_command()
    @maybe_async()
    def publish_to_channel(
        self,
        channel_name: str,
        message: dict | str,
        record_log: bool = True,
        max_age: int = None,
    ) -> bool:
        """Publish a message to a channel.

        .. note:: This method can be called in both sync and asynchronous contexts.

        There is a special case with the `max_age` parameter where you can set it to `-1` to publish the message immediately.
        This is useful for ensuring that the message is sent to the cloud immediately and won't be bundled in with other messages.
        A good example is when you want to clear the UI. However, use this sparingly!

        Examples
        --------

        Simple example::

        >>> result = await self.device_agent.publish_to_channel("my_channel", {"foo": "bar"})
        >>> print(f"Publishing the message was {'successful' if result else 'unsuccessful'}.")


        Publish to the cloud in 5 seconds::

        >>> result = await self.device_agent.publish_to_channel("my_channel", {"foo": "bar"}, max_age=5)
        >>> print(f"Publishing the message was {'successful' if result else 'unsuccessful'}.")

        Publish to the cloud immediately and save to the log::

        >>> result = await self.device_agent.publish_to_channel("my_channel", {"foo": "bar"}, record_log=True, max_age=-1)
        >>> print(f"Publishing the message was {'successful' if result else 'unsuccessful'}.")



        Parameters
        ----------
        channel_name : str
            Name of channel to publish data too.
        message : dict or str
            The data to send either in a dictionary or string format
        record_log :
            Whether to save to the log
        max_age : int
            The maximum age of the message before publishing to the cloud

        Returns
        -------
        bool
            Whether publishing was successful
        """
        if isinstance(message, dict):
            message = json.dumps(message)

        req = device_agent_pb2.ChannelWriteRequest(
            header=device_agent_pb2.RequestHeader(app_id=self.app_key),
            channel_name=channel_name,
            message_payload=message,
            save_log=record_log,
            max_age=max_age,
        )
        resp = self.make_request("WriteToChannel", req)
        return resp and resp.response_header.success or False

    async def publish_to_channel_async(
        self,
        channel_name: str,
        message: dict | str,
        record_log: bool = True,
        max_age: int = None,
    ):
        if isinstance(message, dict):
            message = json.dumps(message)

        req = device_agent_pb2.ChannelWriteRequest(
            header=device_agent_pb2.RequestHeader(app_id=self.app_key),
            channel_name=channel_name,
            message_payload=message,
            save_log=record_log,
            max_age=max_age,
        )
        resp = await self.make_request_async("WriteToChannel", req)
        return resp and resp.response_header.success or False

    @staticmethod
    def _parse_get_token_response(
        response,
    ) -> tuple[str, datetime.datetime, str] | None:
        if response is None:
            return None

        try:
            return (
                response.token,
                datetime.datetime.fromtimestamp(float(response.valid_until)),
                response.endpoint,
            )
        except (ValueError, Exception) as e:
            logging.error("Failed to parse output from get_temp_token", exc_info=e)
            return None

    @cli_command()
    @maybe_async()
    def get_temp_token(self) -> tuple[str, datetime.datetime, str] | None:
        """Get a temporary API token.

        .. deprecated:: 0.4.0
            WSS Tokens are now valid for API calls. Use that instead.

        .. note:: This method can be called in both sync and asynchronous contexts.

        Returns
        -------
        tuple, optional
            (token, expire_time, url_endpoint) if the request succeeds, otherwise None.
        """
        resp = self.make_request(
            "GetTempAPIToken", device_agent_pb2.TempAPITokenRequest()
        )
        return self._parse_get_token_response(resp)

    async def get_temp_token_async(
        self,
    ) -> tuple[str, datetime.datetime, str] | None:
        resp = await self.make_request_async(
            "GetTempAPIToken", device_agent_pb2.TempAPITokenRequest()
        )
        return self._parse_get_token_response(resp)

    async def close(self):
        for listener in self._listeners.values():
            listener.cancel()
        logging.info("Closing device agent interface...")

    def test_dda_available(self):
        try:
            self.test_comms()
            return True
        except Exception as e:
            log.error(f"Failed to get DDA comms: {e}")
            return False

    @cli_command()
    def test_comms(self, message: str = "Comms Check Message") -> str | None:
        """Test connection by sending a basic echo response to device agent container.

        Parameters
        ----------
        message : str
            Message to send to device agent to have echo'd as a response

        Returns
        -------
        str
            The response from device agent.
        """
        return self.make_request(
            "TestComms",
            device_agent_pb2.TestCommsRequest(message=message),
            response_field="response",
        )

    @cli_command()
    def listen_channel(self, channel_name: str) -> None:
        """Listen to channel printing the output to the console.

        Parameters
        ----------
        channel_name : str
            Name of channel to get aggregate from.

        Returns
        -------
        None
            Response is printed to stdout directly
        """
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            asyncio.run(self.run_channel_listening(channel_name))
        else:
            asyncio.create_task(self.run_channel_listening(channel_name))

    async def run_channel_listening(self, channel_name: str):
        def callback(name, aggregate):
            print(name, json.dumps(aggregate))
            sys.stdout.flush()

        self.add_subscription(channel_name, callback)

        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            await self.close()


device_agent_iface = DeviceAgentInterface



import logging
import re
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional

from .element import Element
from .misc import Colour, Option, NotSet
from ..utils import call_maybe_async

if TYPE_CHECKING:
    from .manager import UIManager

log = logging.getLogger(__name__)


class Interaction(Element):
    """Base class for all UI interactions.

    Interactions are elements that can be interacted with by the user, such as buttons, sliders, and text inputs.

    Parameters
    ----------
    name: str
        The name of the interaction, used to identify it in the UI.
    display_name: str, optional
        The display name of the interaction, shown in the UI.
    current_value: Any
        The current value of the interaction. Defaults to NotSet.
    default: Any
        The default value for the interaction, used if current_value is NotSet.
    callback: callable, optional
        A callback function that is called when the interaction value changes.
    transform_check: callable, optional
        A function to transform and check the new value before setting it. Defaults to None.
    show_activity: bool, optional
        Whether to show this interaction in the activity log. Defaults to None which uses the site default.
    """

    type = "uiInteraction"

    def __init__(
        self,
        name: str,
        display_name: str = None,
        current_value: Any = NotSet,
        default: Any = None,
        callback=None,
        transform_check=None,
        show_activity: bool | None = None,
        **kwargs,
    ):
        super().__init__(name, display_name, **kwargs)
        self._current_value = current_value

        if default is not None:
            self._default_value = default
        elif "default_val" in kwargs:
            # for backwards compatibility with old code that used default_val
            self._default_value = kwargs.pop("default_val", None)
        else:
            self._default_value = None

        self._manager: Optional["UIManager"] = None

        if callback:
            self.callback = callback
        if transform_check:
            self.transform_check = transform_check

        ## Handle default value 'statelessly' now in current_value property instead of here
        # if self._current_value in (None, NotSet) and self._default_value is not None:
        #     print(f"Coercing {self.name} to default value {self._default_value}")
        #     self.coerce(self._default_value)

        self.show_activity = show_activity

    @property
    def current_value(self):
        """Returns the current value of the interaction."""
        if self._current_value is NotSet and self._default_value is not None:
            return self._default_value
        return self._current_value if self._current_value is not NotSet else None

    @current_value.setter
    def current_value(self, new_val):
        """Sets the current value of the interaction.

        This will also convert datetime objects to epoch seconds for internal storage.
        """
        ## Store all datetime objects as epoch seconds internally
        if isinstance(new_val, datetime):
            new_val = int(new_val.timestamp())
        self._current_value = new_val

    def _json_safe_current_value(self):
        result = self.current_value
        if isinstance(result, datetime):
            return int(result.timestamp())
        if isinstance(result, (set, tuple, list)):
            # fixme: the site currently treats sets and tuples as lists, so lets just give in to that for now.
            return list(result)
        return result

    def callback(self, new_value: Any):
        """Callback function to handle changes to the interaction value.

        Parameters
        ----------
        new_value: Any
            The new value of the interaction.
        """
        return

    def transform_check(self, new_value: Any) -> Any:
        """Transform and check a new value before setting it.

        This is called before any callback functions are invoked.

        By default, this will replace any None values with a set default value.

        Parameters
        ----------
        new_value: Any
            The new value to transform and check.

        Returns
        -------
        Any: The transformed value, or the default value if new_value is None.
        """
        if new_value is None and self._default_value is not None:
            return self._default_value
        else:
            return new_value

    def _is_new_value(self, new_value: Any) -> bool:
        return new_value != self.current_value

    def _handle_new_value_common(self, new_value: Any):
        try:
            new_value = self.transform_check(new_value)
        except Exception as e:
            log.error(f"Error transforming value for {self.name}: {e}")
            return

        log.debug(f"updating new value: {self.name} {new_value}")
        self.current_value = new_value

    def _handle_new_value(self, new_value: Any):
        self._handle_new_value_common(new_value)
        try:
            self.callback(new_value)
        except Exception as e:
            log.error(f"Error in callback for {self.name}: {e}")

    async def _handle_new_value_async(self, new_value: Any):
        self._handle_new_value_common(new_value)
        await call_maybe_async(self.callback, new_value)
        # try:
        #     self.callback(new_value)
        # except Exception as e:
        #     log.error(f"Error in callback for {self.name}: {e}")

    def coerce(self, value: Any, critical: bool = False) -> None:
        """Coerce the interaction to a new value.

        This will update the current value on the site.

        The callback will be invoked once the site has set the value.

        Parameters
        ----------
        value: Any
            The new value to set for the interaction.
        critical: bool
            If True, this interaction is considered critical and will mark the manager as having a pending critical interaction.
            This is used to ensure that critical interactions are processed immediately and logged on the site.

        Returns
        -------
        None
        """
        if critical and self._manager and value != self.current_value:
            self._manager._has_critical_interaction_pending = True

        if self._manager is not None and self._current_value != value:
            # we need a way to keep track of which interactions our application has "changed"
            # so we don't override ui_cmds that other apps set. ui_cmds ideally should be namespaced
            # so this isn't a problem.
            self._manager._changed_interactions.add(self.name)

        self.current_value = value

    def to_dict(self):
        res = super().to_dict()
        if self._current_value is not NotSet:
            res["currentValue"] = self._json_safe_current_value()
        if self.show_activity is not None:
            res["showActivity"] = self.show_activity
        return res


class Action(Interaction):
    """Represents a UI action (button).

    Parameters
    ----------
    name: str
        The name of the action, used to identify it in the UI.
    display_name: str
        The display name of the action, shown in the UI.
    colour: Colour
        The colour of the action button. Defaults to Colour.blue.
    requires_confirm: bool
        Whether the action requires confirmation before execution on the site. Defaults to True.
    disabled: bool
        Whether the button appears disabled in the UI. Defaults to False.
    """

    type = "uiAction"

    def __init__(
        self,
        name: str,
        display_name: str,
        colour: Colour = Colour.blue,
        requires_confirm: bool = True,
        disabled: bool = False,
        **kwargs,
    ):
        super().__init__(name, display_name, **kwargs)
        self.colour = colour
        self.requires_confirm = requires_confirm
        self.disabled = disabled

    def to_dict(self):
        result = super().to_dict()
        result["colour"] = str(self.colour)
        result["requiresConfirm"] = self.requires_confirm
        result["disabled"] = self.disabled
        return result


class WarningIndicator(Interaction):
    """Represents a warning indicator in the UI.

    Parameters
    ----------
    name: str
        The name of the warning indicator, used to identify it in the UI.
    display_name: str
        The display name of the warning indicator, shown in the UI.
    can_cancel: bool
        Whether the warning can be cancelled by the user. Defaults to True.
    """

    type = "uiWarningIndicator"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.can_cancel = kwargs.pop("can_cancel", True)

    def to_dict(self):
        result = super().to_dict()
        result["can_cancel"] = self.can_cancel
        return result


class HiddenValue(Interaction):
    type = "uiHiddenValue"

    def __init__(self, name, **kwargs):
        super().__init__(name, display_name=None, **kwargs)

    def to_dict(self):
        return {
            "name": self.name,
            "type": self.type,
        }


class SlimCommand(Interaction):
    """Base class for UI commands."""

    type = "uiStateCommand"


class StateCommand(SlimCommand):
    """Represents a state command in the UI (ie. a dropdown that the user can select a state)

    Parameters
    ----------
    name: str
        The name of the state command, used to identify it in the UI.
    display_name: str, optional
        The display name of the state command, shown in the UI.
    user_options: list[Option]
        A list of options that the user can select from. Each option is an instance of the Option class.
        If not provided, an empty list will be created.
    """

    type = "uiStateCommand"

    def __init__(
        self,
        name: str,
        display_name: str = None,
        user_options: list[Option] = None,
        **kwargs,
    ):
        super().__init__(name, display_name, **kwargs)

        # A list of doover_ui_element's
        self.user_options = []
        self.add_user_options(*user_options)

    def to_dict(self):
        result = super().to_dict()
        result["userOptions"] = {o.name: o.to_dict() for o in self.user_options}
        return result

    def add_user_options(self, *option: Option):
        """Add user options to the state command.

        Parameters
        ----------
        *option: Option
            One or more Option instances to add to the state command.
        """
        for o in option:
            # still support legacy dict passing of option values.
            if isinstance(o, Option):
                self.user_options.append(o)
            elif isinstance(o, dict):
                self.user_options.append(Option.from_dict(o))

    add_user_option = add_user_options


class Slider(Interaction):
    """Represents a slider in the UI.

    Parameters
    ----------
    name: str
        The name of the slider, used to identify it in the UI.
    display_name: str, optional
        The display name of the slider, shown in the UI.
    min_val: int
        The minimum value of the slider. Defaults to 0.
    max_val: int
        The maximum value of the slider. Defaults to 100.
    step_size: float
        The step size of the slider. Defaults to 0.1.
    dual_slider: bool
        Whether the slider has a dual handle. Defaults to True.
    inverted: bool
        Whether the slider is inverted (i.e., moves left for higher values). Defaults to True.
    icon: str, optional
        An optional icon to display with the slider.
    colours: str, optional
        An optional string representing colours for the slider, such as "red,green,blue".
    """

    type = "uiSlider"

    def __init__(
        self,
        name: str,
        display_name: str = None,
        min_val: int = 0,
        max_val: int = 100,
        step_size: float = 0.1,
        dual_slider: bool = True,
        inverted: bool = True,
        icon: Optional[str] = None,
        colours: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, display_name, **kwargs)
        self.min_val = min_val
        self.max_val = max_val
        self.step_size = step_size
        self.dual_slider = dual_slider
        self.inverted = inverted
        self.icon = icon
        self.colours = colours

    def to_dict(self):
        result = super().to_dict()
        result["min"] = self.min_val
        result["max"] = self.max_val
        result["stepSize"] = self.step_size
        result["dualSlider"] = self.dual_slider
        result["isInverted"] = self.inverted

        # don't pass values if they have a default value of None since we treat this as "remove element" in a diff.
        if self.icon:
            result["icon"] = self.icon
        if self.colours:
            result["colours"] = self.colours

        return result


def action(
    name: str,
    display_name: str = None,
    colour: Colour = Colour.blue,
    requires_confirm: bool = True,
    **kwargs,
):
    """Decorator to mark a function as a UI action.

    This decorator will add the function to the UI as an action button.

    Parameters
    ----------
    name: str
        The name of the action, used to identify it in the UI.
    display_name: str, optional
        The display name of the action, shown in the UI.
    colour: Colour, optional
        The colour of the action button. Defaults to Colour.blue.
    requires_confirm: bool, optional
        Whether the action requires confirmation before execution on the site. Defaults to True.
    """

    def decorator(func):
        func._ui_type = Action
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "colour": colour,
            "requires_confirm": requires_confirm,
            **kwargs,
        }
        return func

    return decorator


def warning_indicator(
    name: str, display_name: str = None, can_cancel: bool = True, **kwargs
):
    """Decorator to mark a function as a UI warning indicator.

    Parameters
    ----------
    name: str
        The name of the warning indicator, used to identify it in the UI.
    display_name: str, optional
        The display name of the warning indicator, shown in the UI.
    can_cancel: bool
        Whether the warning can be cancelled by the user. Defaults to True.
    """

    def decorator(func):
        func._ui_type = WarningIndicator
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "can_cancel": can_cancel,
            **kwargs,
        }
        return func

    return decorator


def state_command(
    name: str, display_name: str = None, user_options: list[Option] = None, **kwargs
):
    """Decorator to mark a function as a UI state command.

    The decorated function will be added to the UI as a state command,
    which allows the user to select a state from a dropdown menu.
    The associated function to this decorator will be called whenever the state command is selected
    and should take one parameter, `new_value`, which is the new value that has been set for this state command.

    Examples
    --------

    A basic state command ::

        @ui.state_command(
            "my_state_command",
            display_name="My State Command",
            user_options=[
                ui.Option("option1", "Option 1"),
                ui.Option("option2", "Option 2")
            ]
        )
        def my_state_command(self, new_value):
            # This function will be called whenever the state command is selected.
            pass


    Parameters
    ----------
    name: str
        The name of the state command, used to identify it in the UI.
    display_name: str, optional
        The display name of the state command, shown in the UI.
    user_options: list[Option], optional
        A list of options that the user can select from. Each option is an instance of the Option class.
        If not provided, an empty list will be created.
    """

    def decorator(func):
        func._ui_type = StateCommand
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "user_options": user_options,
            **kwargs,
        }
        return func

    return decorator


def hidden_value(name: str, **kwargs):
    """Decorator to mark a function as a UI hidden value.

    This decorator will add the function to the UI as a hidden value, which can be used to store
    values that should not be displayed to the user but can be accessed programmatically.

    The associated function to this decorator will be called whenever this value changes
    and should take one parameter, `new_value`,
    which is the new value that has been set for this hidden value.

    Examples
    --------

    A basic hidden value ::

        @ui.hidden_value("my_hidden_value")
        def my_hidden_value(self, new_value):
            # This function will be called whenever the hidden value changes.
            pass


    Parameters
    ----------
    name: str
        The name of the hidden value.
    """

    def decorator(func):
        func._ui_type = HiddenValue
        func._ui_kwargs = {
            "name": name,
            **kwargs,
        }
        return func

    return decorator


def slider(
    name: str,
    display_name: str = None,
    min_val: int = 0,
    max_val: int = 100,
    step_size: float = 0.1,
    dual_slider: bool = True,
    inverted: bool = True,
    icon: Optional[str] = None,
    **kwargs,
):
    """Decorator to mark a function as a UI slider.

    This decorator will add the function to the UI as a slider element.

    The associated function to this decorator will be called whenever the slider value changes
    and should take one parameter, `new_value`, which is the new value that has been set for this slider.
    Examples
    --------

    A basic slider from 0-100 with a step size of 5 ::

        @ui.slider(
            "my_slider",
            display_name="My Slider",
            min_val=0,
            max_val=100,
            step_size=5
        )
        def my_slider(self, new_value):
            # This function will be called whenever the slider value changes.
            pass

    """

    def decorator(func):
        func._ui_type = Slider
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "min_val": min_val,
            "max_val": max_val,
            "step_size": step_size,
            "dual_slider": dual_slider,
            "inverted": inverted,
            "icon": icon,
            **kwargs,
        }
        return func

    return decorator


def callback(pattern: str | re.Pattern[str], global_interaction: bool = False):
    r"""Decorator to mark a function as a UI callback.

    This accepts either a string or a compiled regex expression to match against the UI element name.

    Examples
    --------

    A callback for a single UI element ::

        @ui.callback("di_fetch")
        def fetch_di(self, element, new_value):
            pass


    A generic callback to match multiple elements ::

        @ui.callback(re.compile(r"do_\d+_toggle"))
        def do_toggle(self, element, new_value):
            pass

    .. note::

        Due to how apps namespace interaction names, if you pass in a string, `{app_name}_` will be prepended to the pattern.
        This means that if you want to trigger a callback for a "global" interaction
        (the most notable example being the camera "Get Now" button), global_interaction=True.

    An example global command::

        @ui.callback("camera_get_now", global_interaction=True)
        def camera_get_now(self, element, new_value):
            pass


    .. note::

        Similar to the above, regex patterns will also be re-compiled and prepended with `{app_name}_` by default.
        You can disable this behaviour by passing `global_interaction=True` to the decorator.

    An example regex compiled global command::

        @ui.callback("^test_global_param_\d+$", global_interaction=True)
        def test_global_param(self, element, new_value):
            pass


    Parameters
    ----------
    pattern: str | re.Pattern[str]
        A string or compiled regex pattern that matches the name of the UI element this callback is for.
    global_interaction: bool
        See above examples for when to use this. This is a special case.
    """

    def decorator(func):
        func._is_ui_callback = True
        func._is_ui_global_interaction = global_interaction
        func._ui_callback_pattern = pattern
        return func

    return decorator



from .agent import Agent as Agent
from .channel import Channel as Channel, Processor as Processor, Task as Task
from .client import Client as Client
from .config import ConfigManager as ConfigManager, ConfigEntry as ConfigEntry
from .message import Message as Message
from .exceptions import (
    Forbidden as Forbidden,
    HTTPException as HTTPException,
    NotFound as NotFound,
    DooverException as DooverException,
)



import asyncio
import logging
from collections.abc import Coroutine, Callable

import grpc

from .config import ModbusConfig, ModbusType, ManyModbusConfig
from .grpc_stubs import modbus_iface_pb2, modbus_iface_pb2_grpc
from ..grpc_interface import GRPCInterface
from ...utils import call_maybe_async, maybe_async
from ...cli.decorators import command as cli_command, ignore_alias
from ...config import Schema

log = logging.getLogger(__name__)
ReadRegisterSubscriptionCallback = (
    Callable[[list[int]], None] | Coroutine[[list[int]], None]
)


def two_words_to_32bit_float(word1: int, word2: int, swap: bool = False):
    """Convert two 16-bit words to a 32-bit float."""
    if swap:
        word1, word2 = word2, word1
    return word1 + (word2 * 65536)


class ModbusInterface(GRPCInterface):
    """ModbusInterface is a gRPC interface for interacting with modbus devices.

    It allows for opening and closing modbus buses, reading and writing registers, and subscribing to register updates.
    It is designed to be used with the modbus_iface gRPC service.

    It supports both synchronous and asynchronous operations, depending on the context of the application.

    Attributes
    ----------
    config : Schema
        Configuration schema for the modbus interface, containing modbus bus definitions.
        This is loaded from application config automatically and should be specified in your `app_config.py` file.
    """

    stub = modbus_iface_pb2_grpc.modbusIfaceStub

    def __init__(
        self,
        app_key: str,
        modbus_uri: str = "127.0.0.1:50054",
        is_async: bool = None,
        config: Schema = None,
    ):
        super().__init__(app_key, modbus_uri, is_async)

        self.subscription_tasks = []

        self.config = config
        self.config_complete = False

    async def setup(self):
        self.config_complete = False
        try:
            config = self.config.modbus_config
        except AttributeError:
            log.info("No modbus interfaces defined in config")
            self.config_complete = True
            return

        if isinstance(config, ModbusConfig):
            elems: list[ModbusConfig] = [config]
        elif isinstance(config, ManyModbusConfig):
            elems: list[ModbusConfig] = config.elements
        else:
            log.warning(f"Unsupported modbus config type: {type(config)}")
            self.config_complete = True
            return

        for elem in elems:
            log.info(f"Setting up modbus bus: {elem}")
            if elem.type.value == ModbusType.SERIAL:
                await self.open_bus_async(
                    elem.type.value,
                    elem.name.value,
                    elem.serial_port.value,
                    elem.serial_baud.value,
                    elem.serial_method.value,
                    elem.serial_bits.value,
                    elem.serial_parity.value,
                    elem.serial_stop.value,
                    elem.serial_timeout.value,
                )
            elif elem.type.value == ModbusType.TCP:
                await self.open_bus_async(
                    elem.type.value,
                    elem.name.value,
                    tcp_uri=elem.tcp_uri.value,
                    tcp_timeout=elem.tcp_timeout.value,
                )
            else:
                log.warning(
                    f"Invalid bus type: {elem.type.value}. Expected 'serial' or 'tcp'."
                )

        self.config_complete = True

    def process_response(self, stub_call: str, response, *args, **kwargs):
        resp = super().process_response(stub_call, response, *args, **kwargs)

        if response.response_header.success:
            log.debug("Response was OK. Nothing to process...")
            return resp

        # only for some of the calls we want to ensure a bus is available (e.g. read/write registers)
        # this will fix it for next run but not the current one...
        try:
            configure_bus = kwargs["configure_bus"]
            bus_id = kwargs["bus_id"]
        except KeyError:
            log.debug(
                "Response was not OK, but no bus_id or configure_bus provided. Nothing to do..."
            )
            return resp

        self.ensure_bus_available(bus_id, response.response_header, configure_bus)
        return resp

    def ensure_bus_available(self, bus_id, response_header, configure: bool = True):
        ## if not config_complete, wait for setup to complete
        if not self.config_complete:
            log.debug("Waiting for modbus setup to complete")
            return False

        ## check the bus status from the response and if the bus does not exist, and configure is True, rerun the setup
        for b in response_header.bus_status:
            if b.bus_id == bus_id:
                return b.open

        log.warning(f"Bus {bus_id} not found in response")
        if configure:
            log.info("Reconfiguring modbus iface")
            _t = asyncio.create_task(self.setup())

        return True

    async def close(self):
        log.info("Closing modbus interface")
        for task in self.subscription_tasks:
            task.cancel()

    @staticmethod
    def _get_bus_request(
        bus_type="serial",
        name="default",
        serial_port="/dev/ttyS0",
        serial_baud=9600,
        serial_method="rtu",
        serial_bits=8,
        serial_parity="N",
        serial_stop=1,
        serial_timeout=0.3,
        tcp_uri="127.0.0.1:5000",
        tcp_timeout=2,
    ):
        if bus_type not in ("serial", "tcp"):
            log.error("Invalid bus type: " + str(bus_type))
            return None

        kwargs = {"bus_id": str(name)}

        if bus_type == "serial":
            kwargs["serial_settings"] = modbus_iface_pb2.serialBusSettings(
                port=serial_port,
                baud=serial_baud,
                modbus_method=serial_method,
                data_bits=serial_bits,
                parity=serial_parity,
                stop_bits=serial_stop,
                timeout=serial_timeout,
            )
        elif bus_type == "tcp":
            ip, port = tcp_uri.split(":")
            kwargs["ethernet_settings"] = modbus_iface_pb2.ethernetBusSettings(
                ip=ip, port=int(port), timeout=tcp_timeout
            )
        else:
            log.error("Invalid bus type: " + str(bus_type))
            return None

        return modbus_iface_pb2.openBusRequest(**kwargs)

    @cli_command()
    @maybe_async()
    def open_bus(
        self,
        bus_type="serial",
        name="default",
        serial_port="/dev/ttyS0",
        serial_baud=9600,
        serial_method="rtu",
        serial_bits=8,
        serial_parity="N",
        serial_stop=1,
        serial_timeout=0.3,
        tcp_uri="127.0.0.1:5000",
        tcp_timeout=2,
    ) -> bool:
        req = self._get_bus_request(
            bus_type,
            name,
            serial_port,
            serial_baud,
            serial_method,
            serial_bits,
            serial_parity,
            serial_stop,
            serial_timeout,
            tcp_uri,
            tcp_timeout,
        )
        if req is None:
            return

        resp = self.make_request("openBus", req)
        return resp.response_header.success

    async def open_bus_async(
        self,
        bus_type="serial",
        name="default",
        serial_port="/dev/ttyS0",
        serial_baud=9600,
        serial_method="rtu",
        serial_bits=8,
        serial_parity="N",
        serial_stop=1,
        serial_timeout=0.3,
        tcp_uri="127.0.0.1:5000",
        tcp_timeout=2,
    ) -> bool:
        req = self._get_bus_request(
            bus_type,
            name,
            serial_port,
            serial_baud,
            serial_method,
            serial_bits,
            serial_parity,
            serial_stop,
            serial_timeout,
            tcp_uri,
            tcp_timeout,
        )
        if req is None:
            return False

        resp = await self.make_request_async("openBus", req)
        return resp.response_header.success

    @cli_command()
    @maybe_async()
    def close_bus(self, bus_id: str = "default") -> bool:
        req = modbus_iface_pb2.closeBusRequest(bus_id=str(bus_id))
        resp = self.make_request("closeBus", req)
        return resp.response_header.success and resp.bus_status.open

    async def close_bus_async(self, bus_id: str = "default") -> bool:
        req = modbus_iface_pb2.closeBusRequest(bus_id=str(bus_id))
        resp = await self.make_request_async("closeBus", req)
        return resp.response_header.success and resp.bus_status.open

    def _validate_read_register_resp(self, resp, bus_id, configure_bus):
        try:
            if not resp.response_header.success:
                log.error("Error reading registers from bus " + str(bus_id))
                return False
            # return self.ensure_bus_availabe(bus_id, resp.response_header, configure_bus)
            return True
        except Exception as e:
            log.error("Error validating read register response: " + str(e))
            return False

    @cli_command()
    @maybe_async()
    def get_bus_status(self, bus_id: str = "default") -> bool:
        """Get the status of a modbus bus.

        .. note:: This method can be called in both sync and asynchronous contexts.

        Parameters
        ----------
        bus_id : str, optional
            The bus ID to fetch an OK status for

        Returns
        -------
        bool
            True if the bus is open, False otherwise.
        """
        req = modbus_iface_pb2.busStatusRequest(bus_id=str(bus_id))
        resp = self.make_request("busStatus", req)
        return resp.response_header.success and resp.bus_status.open

    async def get_bus_status_async(self, bus_id: str = "default") -> bool:
        req = modbus_iface_pb2.busStatusRequest(bus_id=str(bus_id))
        resp = await self.make_request_async("busStatus", req)
        return resp.response_header.success and resp.bus_status.open

    getBusStatus = ignore_alias(get_bus_status)
    getBusStatus_async = get_bus_status_async

    @staticmethod
    def _parse_register_output(values):
        if len(values) == 0:
            return None
        if len(values) == 1:
            return values[0]
        return values

    @cli_command()
    @maybe_async()
    def read_registers(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        num_registers: int = 1,
        register_type: int = 4,
        configure_bus: bool = True,
    ) -> int | list[int] | None:
        """Read registers from a modbus bus.

        .. note:: This method can be called in both sync and asynchronous contexts.

        Examples
        --------
        >>> self.modbus_iface.read_registers(bus_id="default", modbus_id=1, start_address=0, num_registers=10)


        Parameters
        ----------
        bus_id : str, optional
            The bus ID to read registers from (default is "default")
        modbus_id : int, optional
            The modbus ID of the device to read registers from (default is 1)
        start_address : int, optional
            The starting address of the registers to read (default is 0)
        num_registers : int, optional
            The number of registers to read (default is 1)
        register_type : int, optional
            The type of registers to read (default is 4, which is typically holding registers)
        configure_bus : bool, optional
            Whether to configure the bus if it is not available (default is True)

        Returns
        -------
        int | list[int] | None
            The values read from the registers.
            If only one register is read, returns an int.
            If multiple registers are read, returns a list of ints.
            If the response failed, returns None.
        """
        req = modbus_iface_pb2.readRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            count=num_registers,
        )
        resp = self.make_request(
            "readRegisters", req, bus_id=bus_id, configure_bus=configure_bus
        )
        return resp and self._parse_register_output(resp.values)

    async def read_registers_async(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        num_registers: int = 1,
        register_type: int = 4,
        configure_bus: bool = True,
    ) -> int | list[int] | None:
        req = modbus_iface_pb2.readRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            count=num_registers,
        )
        resp = await self.make_request_async(
            "readRegisters", req, bus_id=bus_id, configure_bus=configure_bus
        )
        return resp and self._parse_register_output(resp.values)

    @cli_command()
    @maybe_async()
    def write_registers(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        values: list[int] = None,
        register_type: int = 4,
        configure_bus: bool = True,
    ) -> bool:
        """Write values to registers on a modbus bus.

        .. note:: This method can be called in both sync and asynchronous contexts.

        Examples
        --------
        >>> self.modbus_iface.write_registers(
        ...     bus_id="my_bus",
        ...     modbus_id=1,
        ...     start_address=0,
        ...     values=[100, 200, 300],
        ...     register_type=4,
        ...     configure_bus=True,
        ... )

        Parameters
        ----------
        bus_id: str
            The bus ID to write registers to (default is "default")
        modbus_id: int
            The modbus ID of the device to write registers to (default is 1)
        start_address: int
            The starting address of the registers to write (default is 0)
        values: list[int]
            Register values to write
        register_type: int
            The type of registers to write (default is 4, which is typically holding registers)
        configure_bus: bool
            Whether to configure the bus if it is not available (default is True)

        Returns
        -------
        bool
            True if the write operation was successful, False otherwise.
        """
        values = values or []
        req = modbus_iface_pb2.writeRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            values=values,
        )
        resp = self.make_request(
            "writeRegisters", req, bus_id=bus_id, configure_bus=configure_bus
        )
        return resp and self._validate_read_register_resp(resp, bus_id, configure_bus)

    async def write_registers_async(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        values: list[int] = None,
        register_type: int = 4,
        configure_bus: bool = True,
    ) -> bool:
        values = values or []
        req = modbus_iface_pb2.writeRegisterRequest(
            bus_id=str(bus_id),
            modbus_id=modbus_id,
            register_type=register_type,
            address=start_address,
            values=values,
        )
        resp = await self.make_request_async(
            "writeRegisters", req, bus_id=bus_id, configure_bus=configure_bus
        )
        return resp and self._validate_read_register_resp(resp, bus_id, configure_bus)

    def add_read_register_subscription(
        self,
        bus_id: str = "default",
        modbus_id: int = 1,
        start_address: int = 0,
        num_registers: int = 1,
        register_type: int = 4,
        poll_secs: int = 3,
        callback: ReadRegisterSubscriptionCallback = None,
    ):
        """Add a subscription to read registers from a modbus bus.

        This method creates a subcscription that will periodically read registers from the specified modbus device and
        invoke the provided callback when a read request succeeds.

        The provided callback can be a regular function or a coroutine.

        Examples
        --------

        >>> def my_callback(values: list[int]):
        ...     print("Received new register values:", values)
        >>> self.modbus_iface.add_read_register_subscription(
        ...     bus_id="my_bus",
        ...     modbus_id=1,
        ...     start_address=0,
        ...     num_registers=10,
        ...     callback=my_callback,
        ... )


        Parameters
        ----------
        bus_id : str, optional
            The bus ID to read registers from (default is "default")
        modbus_id : int, optional
            The modbus ID of the device to read registers from (default is 1)
        start_address : int, optional
            The starting address of the registers to read (default is 0)
        num_registers : int, optional
            The number of registers to read (default is 1)
        register_type : int, optional
            The type of registers to read (default is 4, which is typically holding registers)
        poll_secs : int, optional
            The polling interval in seconds for the subscription (default is 3 seconds)
        callback : Callback
            The callback function to invoke when a read request succeeds.
            This accepts a list of integers representing the register values.
            If only one register is read, this will be a single integer.
            This callback can be a regular function or a coroutine.

        """

        if callback is None:
            log.error("No callback provided for read register subscription")
            return None

        try:
            new_task = asyncio.create_task(
                self.run_read_register_subscription_task(
                    bus_id=str(bus_id),
                    modbus_id=modbus_id,
                    start_address=start_address,
                    num_registers=num_registers,
                    register_type=register_type,
                    poll_secs=poll_secs,
                    callback=callback,
                )
            )

            self.subscription_tasks.append(new_task)
            new_task.add_done_callback(self.subscription_tasks.remove)
            return new_task

        except Exception as e:
            log.error("Error adding read register subscription: " + str(e))
            return None

    async def run_read_register_subscription_task(
        self,
        bus_id: str,
        modbus_id: int,
        start_address: int,
        num_registers: int,
        register_type: int,
        poll_secs: int,
        callback: ReadRegisterSubscriptionCallback,
        configure_bus: bool = True,
    ):
        try:
            async with grpc.aio.insecure_channel(self.uri) as channel:
                stub = modbus_iface_pb2_grpc.modbusIfaceStub(channel)
                request = modbus_iface_pb2.readRegisterSubscriptionRequest(
                    bus_id=str(bus_id),
                    modbus_id=modbus_id,
                    register_type=register_type,
                    address=start_address,
                    count=num_registers,
                    poll_secs=poll_secs,
                )

                try:
                    async for response in stub.readRegisterSubscription(request):
                        success = response.response_header.success
                        if not self._validate_read_register_resp(
                            response, bus_id, configure_bus
                        ):
                            values = None
                        elif len(response.values) == 1:
                            values = response.values[0]
                        else:
                            values = response.values

                        log.debug(
                            f"Received new modbus subscription result on bus {bus_id}, for modbus_id {modbus_id}, result={success}"
                        )
                        if callback is not None:
                            await call_maybe_async(callback, values)

                except Exception as e:
                    log.error("Error in read register subscription task: " + str(e))
                    return None

        except Exception as e:
            log.error("Error in read register subscription task: " + str(e))
            return None

    @cli_command()
    def test_comms(self, message: str = "Comms Check Message") -> str | None:
        """Test connection by sending a basic echo response to modbus interface container.

        Parameters
        ----------
        message : str
            Message to send to modbus interface to have echo'd as a response

        Returns
        -------
        str
            The response from modbus interface.
        """
        return self.make_request(
            "testComms",
            modbus_iface_pb2.testCommsRequest(message=message),
            response_field="response",
        )


modbus_iface = ModbusInterface



#!/usr/bin/env python3

import time
import logging
import asyncio

## A generic alarm class that can be used to trigger things via a callback function when a threshold is met
## threshold can be greater than or less than a specified value

## Uses grace_period and min_inter_alarm to prevent rapid triggering of the alarm
## If threshold is met, the alarm will trigger and the callback will be called

## Grace period is the amount of time which the threshold has to be met before the alarm is triggered again
## Min inter alarm is the minimum time between alarms
log = logging.getLogger(__name__)


class Alarm:
    def __init__(
        self,
        threshold_met,
        callback=None,
        grace_period=None,
        min_inter_alarm=None,
    ):
        self.default_grace_period = 60 * 60  # an hour
        self.default_min_inter_alarm = 60 * 60 * 24  # a day

        self.threshold_met = threshold_met
        self.callback = callback
        self.grace_period = grace_period or self.default_grace_period
        self.min_inter_alarm = min_inter_alarm or self.default_min_inter_alarm

        self.last_alarm_time = None
        self.initial_trigger_time = None

    async def check_value(
        self,
        value,
        threshold_met,
        grace_period=None,
        min_inter_alarm=None,
    ):
        if grace_period is not None:
            self.grace_period = grace_period

        if min_inter_alarm is not None:
            self.min_inter_alarm = min_inter_alarm

        if self.threshold_met(value) is False:
            log.debug(f"Threshold not met: {value}")
            self.initial_trigger_time = None
            return False

        else:
            log.debug(f"Threshold met: {value}")
            if self._check_grace_period():
                log.debug(f"Grace period met: {value}")
                if self._check_min_inter_alarm():
                    log.debug(f"Min inter alarm met: {value}")
                    await self._trigger_alarm()
                else:
                    log.debug(f"Min inter alarm not met: {value}")
            else:
                log.debug(f"Grace period not met: {value}")

    def _check_grace_period(self):
        if self.initial_trigger_time is None:
            self.initial_trigger_time = time.time()
            return False

        else:
            if self.initial_trigger_time + self.grace_period < time.time():
                return True
            else:
                return False

    def _check_min_inter_alarm(self):
        if self.last_alarm_time is None:
            return True
        else:
            if self.last_alarm_time + self.min_inter_alarm < time.time():
                return True
            else:
                return False

    async def _trigger_alarm(self):
        if self.callback:
            if asyncio.iscoroutinefunction(self.callback):
                await self.callback()
            else:
                self.callback()
        self.last_alarm_time = time.time()

    def reset_alarm(self):
        self.last_alarm_time = None
        self.initial_trigger_time = None


def create_alarm(
    func,
    threshold_met,
    callback=None,
    grace_period=None,
    min_inter_alarm=None,
):
    """A decorator to check an alarm against the return value of a function
    The function will fire the inputted callback if the alarm is triggered
    See below for an example of how to use this decorator

    Parameters
    ----------
    threshold_met
    callback
    grace_period
    min_inter_alarm

    Returns
    -------

    Example
    -------

        self.get_test_increment = create_alarm(
            self.get_test_increment,
            lambda x:x>20,
            callback=self.test_alarm_callback,
            grace_period=15,
            min_inter_alarm=60,
        )

    async def get_test_increment(self):
        return self.test_increment


    in the above case, the alarm will be checked each time the get_test_increment method is called
    if the value returned is greater than 20 for at least 15 seconds, the callback will be called
    the callback will be called at most once every 60 seconds
    """

    alarm = Alarm(
        threshold_met=threshold_met,
        callback=callback,
        grace_period=grace_period,
        min_inter_alarm=min_inter_alarm,
    )

    async def wrapper(*args, **kwargs):
        result = await func(*args, **kwargs)

        await alarm.check_value(result, threshold_met)

        return result

    wrapper.alarm = alarm

    return wrapper
